#! /usr/bin/env node

/**
 * IBuildInfo interface and typings
 */
interface IBuildInfo {
    hash?: string; // Latest commit hash
    timestamp?: string; // Timestamp on when the build was made
    user?: string; // Current git user
    version?: string; // `version` from package.json
    message?: string; // Custom build message
}

/**
 * Dependencies
 */
import { exec } from "child_process";
import { promises as fs } from "fs";
import moment from "moment";
import prompts from "prompts";
import signale from "signale";
import { promisify } from "util";
const execAsync = promisify(exec);

/**
 * Information sources
 */
// tslint:disable-next-line: no-var-requires
const version = require(`${process.cwd()}/package.json`).version;
const args: string[] = process.argv.slice(2);
const opts: string[] = [
    "--help", // Shows manual
    "--init", // Perform initial setup
    "--no-hash", // Don't show commit hash in build info
    "--no-user", // Don't show git user in build info
    "--no-version", // Don't show package version in build info
    "--no-time", // Don't show timestamp in build info
    "--no-message", // Don't add build message
];

/**
 * Get relevant build information
 */
async function getGitUser(): Promise<string | undefined> {
    const { stdout, stderr } = await execAsync("git config user.name");
    if (stderr) {
        signale.error("Error getting git user, skipping...");
        return undefined;
    } else {
        signale.success("Found git user:", stdout.replace("\n", ""));
        return stdout.replace("\n", "");
    }
}

async function getCommitHash(): Promise<string | undefined> {
    const { stdout, stderr } = await execAsync("git rev-parse --short HEAD");
    if (stderr) {
        signale.error("Error getting latest commit hash, skipping...");
        return undefined;
    } else {
        signale.success("Found commit hash:", stdout.replace("\n", ""));
        return stdout.replace("\n", "");
    }
}

async function getBuildMessage(): Promise<string | undefined> {
    const response = await prompts({
        type: "text",
        name: "message",
        message: "Add build message? (optional)",
    });
    return response.message === "" ? undefined : response.message;
}

async function buildInfo(): Promise<void> {
    signale.start("Collecting build information...");

    // Check provided args if they exist
    args.map((arg: any) => {
        if (!opts.includes(arg)) {
            signale.error(`Error: Unknown arg "${arg}", please re-check arguments before running the tool again!`);
            process.exit(1);
        }
    });

    // Object we store build info in
    const build: IBuildInfo = {};

    if (!args.includes("--no-hash")) {
        build.hash = await getCommitHash();
    }

    if (!args.includes("--no-user")) {
        build.user = await getGitUser();
    }

    if (!args.includes("--no-message")) {
        build.message = await getBuildMessage();
    }

    if (!args.includes("--no-version")) {
        build.version = version;
    }

    if (!args.includes("--no-time")) {
        build.timestamp = moment().format("MMMM DD, YYYY HH:mm:ss");
    }

    signale.info(build);

    // Path we write the file to, should match your Angular projects `src/` folder
    const exportPath = process.cwd() + "/src/build.ts";

    // Try writing file, if it fails, display error message
    try {
        await fs.writeFile(
            exportPath,
            "// Angular build information, automatically generated by `angular-build-info`\n" +
                "export const buildInfo = " +
                JSON.stringify(build, null, 4).replace(/\"([^(\")"]+)\":/g, "$1:") +
                ";\n",
        );
        signale.success(`Saved build information to \`${exportPath}\``);
    } catch {
        signale.error(`An error occured writing to \`${exportPath}\`, does the path exist?`);
    }
}

async function init(): Promise<void> {
    // Path we write the file to, should match your Angular projects `src/` folder
    const exportPath = process.cwd() + "/src/build.ts";

    signale.info("Welcome to `angular-build-info`!");
    signale.info(
        `We will now create a boilerplate \`build.ts\` file in \`${exportPath}\` and fill it with basic information so you can start implementing it in your front-end.`,
    );
    signale.info(
        `If you wish for more info on how to implement the provided info in your Angular app, feel free to check out the main repo over at https://github.com/4dams/angular-build-info`,
    );

    signale.start("Creating `build.ts` file...");

    const template = {
        hash: "1e872b5",
        timestamp: moment().format("MMMM DD, YYYY HH:mm:ss"),
        user: "Octocat",
        version: "1.0.0",
    };

    // Try writing file, if it fails, display error message
    try {
        await fs.writeFile(
            exportPath,
            "// Angular build information, automatically generated by `angular-build-info`\n" +
                "export const buildInfo = " +
                JSON.stringify(template, null, 4).replace(/\"([^(\")"]+)\":/g, "$1:") +
                ";\n",
        );

        signale.success("Successfully created `build.ts` file!");
        signale.info(
            "You should now modify your build/deploy scripts in your `package.json` file to run this script every time before your Angular app is built. An example would be:",
        );
        signale.info(`[...] "build": "build-info && ng build --prod", [...]`);
        signale.info("Again, you can find more info on implementing this tool on the main repo.");
    } catch {
        signale.error(`An error occured writing to \`${exportPath}\`, does the path exist?`);
    }
}

async function displayManual() {
    signale.info("Welcome to `angular-build-info`!");
    signale.info("");
    signale.info("--help          Displays this message");
    signale.info("--init          Creates template `build.ts` file so you can start implementing it");
    signale.info("--no-hash       Will not add latest commit hash to final `build.ts`");
    signale.info("--no-user       Will not add git username to final `build.ts`");
    signale.info("--no-version    Will not add version from `package.json` to `build.ts`");
    signale.info("--no-time       Will not add timestamp to final `build.ts`");
}

/**
 * Main function we run upon starting the script
 */
function start(): void {
    if (args.includes("--init")) {
        init();
        return;
    }

    if (args.includes("--help")) {
        displayManual();
        return;
    }

    buildInfo();
}

start();
